# coding: utf-8
"""Parse and format OSGB grid reference strings.

This module provides functions to parse and format grid references, and
to tell you which maps include a given reference.
"""
from __future__ import unicode_literals, print_function, division

import ast
import collections
import itertools
import math
import pkgutil
import re

__all__ = ['format_grid', 'parse_grid', 'sheet_keys', 'get_sheet']

GRID_SQ_LETTERS = 'VWXYZQRSTULMNOPFGHJKABCDE'
GRID_SIZE = int(math.sqrt(len(GRID_SQ_LETTERS)))
MINOR_GRID_SQ_SIZE = 100000
MAJOR_GRID_SQ_SIZE = GRID_SIZE * MINOR_GRID_SQ_SIZE
MAJOR_GRID_SQ_EASTING_OFFSET = 2 * MAJOR_GRID_SQ_SIZE
MAJOR_GRID_SQ_NORTHING_OFFSET = 1 * MAJOR_GRID_SQ_SIZE
MAX_GRID_SIZE = MINOR_GRID_SQ_SIZE * len(GRID_SQ_LETTERS)


def get_sheet(key):
    '''Fetch a map from the locker:

    >>> print(get_sheet("A:1").title)
    Shetland - Yell, Unst and Fetlar

    '''
    try:
        return map_locker[key]
    except KeyError:
        return None


"""Some checks for the maps

>>> 'A' in name_for_map_series
True

>>> all(map_locker[label].series in name_for_map_series for label in map_locker)
True

>>> all(len(map_locker[label].bbox) == 2 for label in map_locker)
True
"""

name_for_map_series = {
    'A': 'OS Landranger',
    'B': 'OS Explorer',
    'C': 'OS One-Inch 7th series',
    'H': 'Harvey British Mountain Maps',
    'J': 'Harvey Superwalker',
}

Sheet = collections.namedtuple("Sheet", "bbox area series number parent title polygon")
map_locker = {}


def _load_maps(series, filename):
    '''Read a maps file into a dict'''
    maps = dict()
    for line in pkgutil.get_data('osgb', filename).decode('utf-8').splitlines():
        sheet, inset, bbox, area, title, polygon = (x.strip() for x in line.split('  ') if x)
        key = '{}:{}'.format(series, sheet)
        if inset == '.':
            parent = ''
        else:
            parent = key
            key += '.' + inset

        if series == 'B':
            number = sheet.strip('NEWS')
        else:
            number = sheet

        maps[key] = Sheet(
            ast.literal_eval(bbox),
            float(area),
            series, number, parent, title,
            ast.literal_eval(polygon)
        )

    return maps


map_locker.update(_load_maps('A', 'maps-landranger.txt'))
map_locker.update(_load_maps('B', 'maps-explorer.txt'))
map_locker.update(_load_maps('C', 'maps-one-inch.txt'))
map_locker.update(_load_maps('H', 'maps-harvey-mountain.txt'))
map_locker.update(_load_maps('J', 'maps-harvey-superwalker.txt'))


class Error(Exception):
    """Parent class for Gridder exceptions"""
    pass


class GridParseError(Error):
    """Parent class for parsing exceptions"""
    pass


class GridFormatError(Error):
    """Parent class for formatting exceptions"""
    pass


class GarbageError(GridParseError):
    """Raised when no grid ref can be deduced from the string given.

    Attributes:
        input

    """
    def __init__(self, spam):
        self.spam = spam

    def __str__(self):
        return "I can't read a grid reference from this -> {}".format(self.spam)


class SheetMismatchError(GridParseError):
    """Raised when grid ref given is not on sheet given

    Attributes:
        sheet
        easting
        northing

    """
    def __init__(self, sheet, easting, northing):
        self.sheet = sheet
        self.easting = easting
        self.northing = northing

    def __str__(self):
        return "Grid point ({}, {})".format(self.easting, self.northing) \
               + " is not on sheet {}".format(self.sheet)


class UndefinedSheetError(GridParseError):
    """Raised when sheet given is not one we know

    Attributes:
        sheet

    """
    def __init__(self, sheet):
        self.sheet = sheet

    def __str__(self):
        return "Sheet {} is not known here.".format(self.sheet)


class FaultyFormError(GridFormatError):
    """Raised when the form given to format_grid is unmatched.

    Attributes:
       form

    """
    def __init__(self, form):
        self.form = form

    def __str__(self):
        return "This form argument was not matched --> form='{}'".format(self.form)


class FarFarAwayError(GridFormatError):
    """Raised when grid reference is nowhere near GB.

    Attributes:
       northing
       easting
    """
    def __init__(self, easting, northing):
        self.easting = easting
        self.northing = northing

    def __str__(self):
        return "The spot with coordinates " \
            + "({}, {})".format(self.easting, self.northing) \
            + " is too far from the OSGB grid"


def _is_left_right_or_on(x, y, a, b):
    '''Is the point (x, y) left of, right of, or on the line from a to b?
    Left: > 0, On: == 0, Right: < 0

    >>> _is_left_right_or_on(0, 0, (42,-4), (42, +4)) > 0
    True
    >>> _is_left_right_or_on(0, 0, (0, 42), (0, -42)) == 0
    True
    >>> _is_left_right_or_on(0, 0, (-42,-4), (-42, +4)) < 0
    True
    '''
    return (b[0] - a[0]) * (y - a[1]) - (x - a[0]) * (b[1] - a[1])


def _pairwise(iterable):
    # _pairwise('ABCDEFG') --> AB BC CD DE EF FG
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def _winding_number(x, y, poly):
    '''This is adapted from http://geomalgorithms.com/a03-_inclusion.html'''
    w = 0
    for a, b in _pairwise(poly):
        if a[1] <= y:
            if b[1] > y and _is_left_right_or_on(x, y, a, b) > 0:
                w += 1
        else:
            if b[1] <= y and _is_left_right_or_on(x, y, a, b) < 0:
                w -= 1
    return w


def format_grid(easting, northing=None, form='SS EEE NNN'):
    """Formats an (easting, northing) pair into traditional grid reference.

    This routine formats an (easting, northing) pair into a traditional
    grid reference with two letters and two sets of three numbers, like this:
    ``SU 387 147``.

    >>> print(format_grid(438710.908, 114792.248))
    SU 387 147

    If you want the individual components, apply ``split()`` to it.

    >>> print('-'.join(format_grid(438710.908, 114792.248).split()))
    SU-387-147

    and note that the results are strings not integers.  Note also that
    rather than being rounded, the easting and northing are truncated
    (as the OS system demands), so the grid reference refers to the
    lower left corner of the relevant square.  The system is described
    below the legend on all OS Landranger maps.

    The ``format_grid`` routine takes an optional keyword argument
    ``form``, that controls the form of grid reference returned.

    >>> print(format_grid(438710.908, 114792.248, form='SS EEE NNN'))
    SU 387 147
    >>> print(format_grid(438710.908, 114792.248, form='SS EEEEE NNNNN'))
    SU 38710 14792
    >>> print(format_grid(438710.908, 114792.248, form='SS'))
    SU
    >>> print(format_grid(438710.908, 114792.248, form='SSEN'))
    SU31
    >>> print(format_grid(438710.908, 114792.248, form='SSEENN'))
    SU3814
    >>> print(format_grid(438710.908, 114792.248, form='SSEEENNN'))
    SU387147
    >>> print(format_grid(438710.908, 114792.248, form='SSEEEENNNN'))
    SU38711479
    >>> print(format_grid(438710.908, 114792.248, form='SSEEEEENNNNN'))
    SU3871014792
    >>> print(format_grid(438710.908, 114792.248, form='SS EN'))
    SU 31
    >>> print(format_grid(438710.908, 114792.248, form='SS EE NN'))
    SU 38 14
    >>> print(format_grid(438710.908, 114792.248, form='SS EEE NNN'))
    SU 387 147
    >>> print(format_grid(438710.908, 114792.248, form='SS EEEE NNNN'))
    SU 3871 1479
    >>> print(format_grid(400010.908, 114792.248, form='SS EEEEE NNNNN'))
    SU 00010 14792

    You can't leave out the SS, you can't have N before E, and there
    must be the same number of Es and Ns.  Except for the two special
    formats:

    >>> print(format_grid(438710.908, 114792.248, form='TRAD'))
    SU 387 147
    >>> print(format_grid(438710.908, 114792.248, form='GPS'))
    SU 38710 14792

    The format can be given as upper case or lower case or a mixture.

    >>> print(format_grid(438710.908, 114792.248, form='trad'))
    SU 387 147

    but in general the form argument must match ``"SS E* N*"`` (spaces optional)

    >>> format_grid(432800, 250000, form='TT') # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    FaultyFormError: This form argument was not matched --> form='TT'

    Here are some more extreme examples:

    >>> print(format_grid(314159, 271828, form='SS'))
    SO
    >>> print(format_grid(0, 0, form='SS'))
    SV
    >>> print(format_grid(432800, 1250000, form='SS'))
    HP

    The arguments can be negative...

    >>> print(format_grid(-5, -5, form='SS'))
    WE

    ...but must not be too far away from the grid:

    >>> format_grid(-1e12, -5) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    FarFarAwayError: The spot with coordinates (-1e+12, -5) is too far from the OSGB grid


    """
    if northing is None:
        (easting, northing) = easting

    e = easting + MAJOR_GRID_SQ_EASTING_OFFSET
    n = northing + MAJOR_GRID_SQ_NORTHING_OFFSET

    if 0 <= e < MAX_GRID_SIZE and 0 <= n < MAX_GRID_SIZE:
        major_index = int(e / MAJOR_GRID_SQ_SIZE) + GRID_SIZE * int(n / MAJOR_GRID_SQ_SIZE)
        e = e % MAJOR_GRID_SQ_SIZE
        n = n % MAJOR_GRID_SQ_SIZE
        minor_index = int(e / MINOR_GRID_SQ_SIZE) + GRID_SIZE * int(n / MINOR_GRID_SQ_SIZE)
        sq = GRID_SQ_LETTERS[major_index] + GRID_SQ_LETTERS[minor_index]

    else:
        raise FarFarAwayError(easting, northing)

    e = int(easting % MINOR_GRID_SQ_SIZE)
    n = int(northing % MINOR_GRID_SQ_SIZE)

    # special cases
    ff = form.upper()
    if ff == 'TRAD':
        ff = 'SS EEE NNN'
    elif ff == 'GPS':
        ff = 'SS EEEEE NNNNN'
    elif ff == 'SS':
        return sq

    m = re.match(r'S{1,2}(\s*)(E{1,5})(\s*)(N{1,5})', ff)
    if m is None:
        raise FaultyFormError(form)

    (space_a, e_spec, space_b, n_spec) = m.group(1, 2, 3, 4)
    e = int(e / 10 ** (5 - len(e_spec)))
    n = int(n / 10 ** (5 - len(n_spec)))

    return sq \
        + space_a + '{0:0{1}d}'.format(e, len(e_spec)) \
        + space_b + '{0:0{1}d}'.format(n, len(n_spec))


def parse_grid(*grid_elements, **kwargs):
    """Parse a grid reference from a range of inputs.

The parse_grid routine extracts a (easting, northing) pair from a
string, or a list of arguments, representing a grid reference.  The pair
returned are in units of metres from the false origin of the grid.

The arguments should be in one of the following three forms

- A single string representing a grid reference

    >>> parse_grid("TA 123 678")
    (512300, 467800)
    >>> parse_grid("TA 12345 67890")
    (512345, 467890)

    The string can also refer to 100km, 10km, 1km, or even 10m squares:

    >>> parse_grid('TA')
    (500000, 400000)
    >>> parse_grid('TA15')
    (510000, 450000)
    >>> parse_grid('TA 12 56')
    (512000, 456000)
    >>> parse_grid('TA 1234 5678')
    (512340, 456780)

    The spaces are optional in all cases:

    >>> parse_grid(" TA 123 678 ")
    (512300, 467800)
    >>> parse_grid(" TA123 678 ")
    (512300, 467800)
    >>> parse_grid(" TA 123678")
    (512300, 467800)
    >>> parse_grid("TA123678")
    (512300, 467800)
    >>> parse_grid("TA1234567890")
    (512345, 467890)

    Here are some more extreme examples:

    >>> parse_grid('SV9055710820') # St Marys lifeboat station
    (90557, 10820)
    >>> parse_grid('HU4795841283') # Lerwick lifeboat station
    (447958, 1141283)
    >>> parse_grid('WE950950') # At sea, off the Scillies
    (-5000, -5000)

    Note in the last one that we are "off" the grid proper.  This lets
    you work with "pseudo-grid-references" like these:

    >>> parse_grid('XD 61191 50692') # St Peter Port the Channel Islands
    (361191, -49308)
    >>> parse_grid('MC 03581 16564') # Rockall
    (-296419, 916564)


- A two or three element list representing a grid reference

    >>> parse_grid('TA', 0, 0)
    (500000, 400000)
    >>> parse_grid('TA', 123, 678)
    (512300, 467800)
    >>> parse_grid('TA', 12345, 67890)
    (512345, 467890)
    >>> parse_grid('TA', '123 678')
    (512300, 467800)
    >>> parse_grid('TA', '12345 67890')
    (512345, 467890)
    >>> parse_grid('TA', '1234567890')
    (512345, 467890)

    Or even just two numbers (primarily included for testing purposes).
    Note that this allows floats, and that the results will come back as
    floats

    >>> parse_grid(314159, 271828)
    (314159.0, 271828.0)
    >>> parse_grid('314159 271828')
    (314159.0, 271828.0)
    >>> parse_grid(231413.123, 802143.456)
    (231413.123, 802143.456)

    If you are processing grid references from some external data source
    beware that if you use a list with bare numbers you may lose any
    leading zeros for grid references close to the SW corner of a grid
    square.  This can lead to some ambiguity.  Either make the numbers
    into strings to preserve the leading digits or supply a keyword
    argument ``figs`` to define how many figures are supposed to be in
    each easting and northing.  Like this:

    >>> parse_grid('TA', 123, 8)
    (512300, 400800)
    >>> parse_grid('TA', 123, 81, figs=5)
    (500123, 400081)

    The default setting of figs is 3, which assumes you are using
    hectometres as in a traditional grid reference. The maximum is 5 and
    the minimum is the length of the longer of easting or northing.

- A string or a list representing a map and a local grid reference,
  corresponding to the following examples:

    >>> parse_grid('176/224711') # Caesar's Camp
    (522400, 171100)
    >>> parse_grid(176, 224, 711)
    (522400, 171100)
    >>> parse_grid('A:164/352194') # Charlbury Station
    (435200, 219400)
    >>> parse_grid('B:OL43E/914701') # map Chesters Bridge
    (391400, 570100)
    >>> parse_grid('B:OL43E 914 701') # map Chesters Bridge
    (391400, 570100)
    >>> parse_grid('B:OL43E', '914701') # map 2-arg Chesters Bridge
    (391400, 570100)
    >>> parse_grid('B:OL43E', 914, 701) # map 3-arg Chesters Bridge
    (391400, 570100)
    >>> parse_grid(164, 513, 62) # Carfax
    (451300, 206200)
    >>> parse_grid('B:119/480103') # map with dual name
    (448000, 110300)
    >>> parse_grid('B:OL3/480103') # map with dual name
    (448000, 110300)
    >>> parse_grid('B:309S.a 26432 34013') # inset on B:309
    (226432, 534013)
    >>> parse_grid('B:368W', 723, 112) # 3-arg, dual name
    (272300, 711200)
    >>> parse_grid('B:OL47W', 723, 112) # 3-arg, dual name
    (272300, 711200)

    or finally just a sheet name;  this will show the SW corner:

    >>> parse_grid('A:82')
    (195000, 530000)

    with the usual rule about assuming you meant Landrangers:

    >>> parse_grid(161)
    (309000, 205000)

    A map sheet with a grid ref that does not actually coincide will
    raise a SheetMismatchError error

    >>> parse_grid('176/924011') # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    SheetMismatchError: Grid point (592400, 201100) is not on sheet A:176

    A map sheet that does not exist will raise an UndefinedSheetError error

    >>> parse_grid('B:999/924011') # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    UndefinedSheetError: Sheet B:999 is not known here.

If there's no matching input then a GarbageError error is raised.

    >>> parse_grid('Somewhere in London') # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    GarbageError: I can't read a grid reference from this -> Somewhere in London

"""

    try:
        figs = int(kwargs['figs'])
    except (KeyError, ValueError):
        figs = 3

    if len(grid_elements) == 3:
        sq, ee, nn = grid_elements
        figs = min(5, max(figs, len(str(ee)), len(str(nn))))
        grid_string = '{0} {1:0{3}d} {2:0{3}d}'.format(sq, int(ee), int(nn), figs)
    else:
        grid_string = ' '.join(str(x).strip() for x in grid_elements)

    # normal case : TQ 123 456 etc
    offsets = _get_grid_square_offsets(grid_string)
    if offsets is not None:
        if len(grid_string) == 2:   # ie must have been just a valid square
            return offsets

        en_tuple = _get_eastings_northings(grid_string)
        if en_tuple is not None:
            return (en_tuple[0] + offsets[0], en_tuple[1] + offsets[1])

    # just a pair of numbers perhaps?
    try:
        grid_elements = tuple(float(x) for x in grid_string.split())
    except ValueError:
        pass
    else:
        if len(grid_elements) == 2:
            return grid_elements

    # probably now a sheet name rather than a SQ number
    # so hand off to the sheet reference parser
    return _get_easting_northing_from_sheet_reference(grid_string)


def _get_easting_northing_from_sheet_reference(possible_map_gr):
    '''Find a grid reference from a sheet number (with optional local GR)

    >>> _get_easting_northing_from_sheet_reference('A:195/291849')
    (429100, 84900)

    >>> _get_easting_northing_from_sheet_reference('A:195/29184')  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    GarbageError: I can't read a grid reference from this -> 29184

    '''

    # so lets try to decompose the string version of the input
    ok = re.match(r'^([A-Z]:)?([0-9NEWSOL/]+?)(\.[a-z]+)?(?:[ -/.]([ 0-9]+))?$', possible_map_gr)
    if not ok:
        raise GarbageError(possible_map_gr)

    (prefix, sheet_number, suffix, numbers) = ok.groups()

    if prefix is None:
        sheet = "A:" + sheet_number  # default to Landranger sheets
    else:
        sheet = prefix + sheet_number

    if suffix is not None:
        sheet = sheet + suffix

    s = get_sheet(sheet)

    if s is None:
        raise UndefinedSheetError(sheet)

    easting = s.bbox[0][0]   # start with SW corner
    northing = s.bbox[0][1]

    if numbers is not None:
        try:
            (e, n) = _get_eastings_northings(numbers)
        except TypeError:  # in case we get None back
            raise GarbageError(numbers)
        easting = easting + (e - easting) % MINOR_GRID_SQ_SIZE
        northing = northing + (n - northing) % MINOR_GRID_SQ_SIZE
        if _winding_number(easting, northing, s.polygon) == 0:
            raise SheetMismatchError(sheet, easting, northing)

    return (easting, northing)


def _get_grid_square_offsets(sq):
    """Get (e, n) for ll corner of a grid square

    >>> _get_grid_square_offsets('SV')
    (0, 0)

    >>> _get_grid_square_offsets('TQ 345 452')
    (500000, 100000)

    >>> _get_grid_square_offsets('') is None
    True

    """

    if len(sq) < 2:
        return None

    a = GRID_SQ_LETTERS.find(sq[0].upper())
    if a < 0:
        return None

    b = GRID_SQ_LETTERS.find(sq[1].upper())
    if b < 0:
        return None

    (Y, X) = divmod(a, GRID_SIZE)
    (y, x) = divmod(b, GRID_SIZE)

    return (
        MAJOR_GRID_SQ_SIZE * X - MAJOR_GRID_SQ_EASTING_OFFSET + MINOR_GRID_SQ_SIZE * x,
        MAJOR_GRID_SQ_SIZE * Y - MAJOR_GRID_SQ_NORTHING_OFFSET + MINOR_GRID_SQ_SIZE * y
    )


def _get_eastings_northings(s):
    """Extract easting and northing from GR string.

    >>> _get_eastings_northings(' 12345 67890')
    (12345, 67890)
    >>> _get_eastings_northings(' 234 567')
    (23400, 56700)
    """
    t = re.findall(r'(\d+)', s)
    if len(t) == 2:
        (e, n) = t
    elif len(t) == 1:
        gr = t[0]
        f = len(gr)
        if f in [2, 4, 6, 8, 10]:
            f = int(f / 2)
            e, n = (gr[:f], gr[f:])
        else:
            return None
    else:
        return None

    figs = min(5, max(len(e), len(n)))
    return (int(e) * 10 ** (5 - figs), int(n) * 10 ** (5 - figs))


def sheet_keys(easting, northing=None, series='ABCHJ'):
    """Return a list of map sheet keys that show the (easting, northing)
    point given.

    The optional argument "series" controls which maps are included in
    the list.  The default is to include maps from all defined series.

    >>> print(' '.join(sheet_keys(438710.908, 114792.248, series='AB')))
    A:196 B:OL22E

    Currently the series included are:

    A:
        OS Landranger 1:50000 maps

    B:
        OS Explorer 1:25000 maps (some of these are designated as "Outdoor Leisure" maps)

    C:
        OS Seventh Series One-Inch 1:63360 maps

    H:
        Harvey British Mountain maps - mainly at 1:40000

    J:
        Harvey Super Walker maps - mainly at 1:25000

    Note that the numbers returned for the Harvey maps have been
    invented for the purposes of this module.  They do not appear on the
    maps themselves; instead the maps have titles.

    You can use the numbers returned as an index to the maps data to
    find the appropriate title.  To get a sheet object use:
    get_sheet(key)

    >>> print(get_sheet(sheet_keys(314159, 271828, series='C')[0]).title)
    Montgomery and Llandrindod Wells

    >>> print(' '.join(sheet_keys(314159, 271828)))
    A:136 A:148 B:200E B:214E C:128

    You can restrict the list to certain series.  So if you only want
    Explorer maps use: series='B', and if you want only Explorers and
    Landrangers use: series='AB', and so on.

    >>> print(''.join(sheet_keys(651537, 313135, series='A')))
    A:134

    If the (easting, northing) pair is not covered by any map sheet
    you'll get an empty list

    >>> sheet_keys(0, 0)
    []

    You can pass a tuple as argument if you prefer, so that you can
    call sheet_keys directly with parse_grid

    >>> print(' '.join(sheet_keys(parse_grid("SZ294849"))))
    A:195 A:196 B:OL29W C:180
    >>> print(' '.join(sheet_keys(parse_grid("SZ294849"), series='A')))
    A:195 A:196

    But we don't call parse_grid automatically for you...

    >>> sheet_keys("SZ294849")
    []

    """
    if northing is None:
        if type(easting) is tuple:
            (easting, northing) = easting
        else:
            return []

    sheets = list()
    for (k, m) in map_locker.items():
        if k[0] not in series:  # initial letter of key is series letter
            continue

        if m.bbox[0][0] <= easting < m.bbox[1][0]:
            if m.bbox[0][1] <= northing < m.bbox[1][1]:
                if _winding_number(easting, northing, m.polygon) != 0:
                    sheets.append(k)

    return sorted(sheets)



"""Conversion between latitude/longitude coordinates and OSGB grid
references.

This module provides the core routines that implement the OS conversion
formulae.

"""
from __future__ import print_function, division, unicode_literals

import array
import math
import pkgutil
import sys

__all__ = ['grid_to_ll', 'll_to_grid']

# The ellipsoid models for projection to and from the grid
# each line has a, b, n, ee
# - a, b are the semi-major and semi-minor axes of the models
# - n = (a-b)/(a+b)
# - ee = 1-(b/a)**2

ELLIPSOID_MODELS = {
    'WGS84': (6378137.000, 6356752.31424518, 0.0016792203863836474, 0.006694379990141108),
    'OSGB36': (6377563.396, 6356256.909, 0.0016732203289874942, 0.006670540074149134),
}

# The defining constants for the OSGB grid
ORIGIN_LAMBDA = -2 / 57.29577951308232087679815481410517
ORIGIN_PHI = 49 / 57.29577951308232087679815481410517
ORIGIN_EASTING = 400000.0
ORIGIN_NORTHING = -100000.0
CONVERGENCE_FACTOR = 0.9996012717

# OSTN data
# Arrays of bytes are handled one way in Python3...
if sys.version_info > (3, 0):
    OSTN_EE_SHIFTS = array.array('H')
    OSTN_EE_SHIFTS.frombytes(pkgutil.get_data("osgb", "ostn_east_shift_82140"))
    OSTN_NN_SHIFTS = array.array('H')
    OSTN_NN_SHIFTS.frombytes(pkgutil.get_data("osgb", "ostn_north_shift_-84180"))

# ... and another in Python2
else:
    OSTN_EE_SHIFTS = array.array(b'H')
    OSTN_EE_SHIFTS.fromstring(pkgutil.get_data("osgb", "ostn_east_shift_82140"))
    OSTN_NN_SHIFTS = array.array(b'H')
    OSTN_NN_SHIFTS.fromstring(pkgutil.get_data("osgb", "ostn_north_shift_-84180"))

OSTN_EE_BASE = 82140
OSTN_NN_BASE = -84180


class Error(Exception):
    """Parent class for exceptions in this module"""
    pass


class UndefinedModelError(Error):
    """Raised when the model given is undefined

    Attributes:
        spam - the name of the unknown model

    """
    def __init__(self, spam):
        self.spam = spam

    def __str__(self):
        return "{}".format(self.spam)


class MissingArgumentError(Error):
    """Raised by grid_to_ll when northing is None and easting is not a tuple

    Attributes:
        easter -- this should have been a tuple

    """
    def __init__(self, easter):
        self.arg = easter

    def __str__(self):
        return "This should have been a tuple: {}".format(self.arg)


def grid_to_ll(easting, northing=None, model='WGS84', rounding=None):
    """
    Convert OSGB (easting, northing) to latitude and longitude.

    Input
        an ``(easting, northing)`` pair in metres from the false point of
        origin of the grid.

        Note that if you are starting with a tradtional grid reference
        string like ``TQ183506``, you need to parse it into an (easting,
        northing) pair using the :py:func:`parse_grid` function from
        :py:mod:`osgb.gridder` before you can pass it to this function::

          lat, lon = osgb.grid_to_ll(osgb.parse_grid("TQ183506"))

    Output
        a ``(latitude, longitude)`` pair in decimal degrees; postive North/East,
        negative South/West (that is, following ISO 6709 conventions).

    An optional argument ``model`` defines the graticule model to use.
    The default is ``WGS84``, the standard model used for the GPS network
    and for references given on Google Earth or Wikipedia, etc.  The
    only other valid value is ``OSGB36`` which is the traditional model
    used in the UK before GPS.  Latitude and longitude coordinates
    marked around the edges of OS maps published before 2015 are given
    in the OSGB36 model.

    An optional argument ``rounding`` controls how many decimal places are
    provided in the output.  Grid references rounded to whole metres will give
    lat/lon that are accurate to about 5 decimal places.  But Grid references
    in millimetres will give results accurate to 8 decimal places.  The default
    value of ``rounding`` is None, and the default behaviour is to round
    appropriately according to the number of decimal places given in the input.

    Effectively if your input is in whole metres you will get 6 places,
    but if your input has any decimal places of metres you will get 9 places.
    You can supply your own value of ``rounding`` if required.  The allowed
    values are the same as the second argument to the Python built-in ``round``
    function.

    Note that more decimal places do not make the conversions more accurate!
    And that in the UK, 0.000001 of a degree of latitude is about 7cm, 0.000001
    of a degree of longitude is about 10cm.

    >>> # Glendessary, the graticule marker on Sheet 33
    >>> grid_to_ll(197574, 794793, model='OSGB36')
    (57.000002, -5.33332)

    >>> # Scorriton
    >>> grid_to_ll(269995, 68361, model='OSGB36', rounding=5)
    (50.5, -3.83333)

    >>> # Cranbourne Chase, on the central meridian
    >>> grid_to_ll(400000, 122350.0439, model='OSGB36')
    (51.0, -2.0)

    >>> # The example from the OSGB documentation
    >>> grid_to_ll(651409.903, 313177.27, model='OSGB36', rounding=8)
    (52.6575703, 1.71792158)

    The routines will produce lots more decimal places, and you can choose
    what rounding you want, although they aren't really meaningful beyond nine
    places, since the conversion routines supplied by the OS are only designed
    to be accurate to about 1mm (8 places).

    >>> # Hoy (Orkney)
    >>> grid_to_ll(323223, 1004000, model='OSGB36', rounding=10)
    (58.9168015046, -3.3333320036)

    >>> # Glen Achcall
    >>> grid_to_ll(217380, 896060, model='OSGB36', rounding=10)
    (57.9167163329, -5.083330214)

    You can provide a tuple instead of two separate args:

    >>> grid_to_ll((217380, 896060))
    (57.916378, -5.084588)

    But you will get an error if the single argument is anything but a tuple

    >>> grid_to_ll("NH 345 213") # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    MissingArgumentError: This should have been a tuple: NH 345 213

    You need to do ``parse_grid`` first to avoid this of course.

    Finally here is an example of how to use the optional keyword arguments:

    >>> grid_to_ll(easting=217380, northing=896060, model='OSGB36')
    (57.916716, -5.08333)

    And a check that conversion works outside the defined OSTN15 polygon:

    >>> grid_to_ll(easting=-100, northing=-100, rounding=8)
    (49.76584553, -7.55843918)

    >>> grid_to_ll(easting=1, northing=1, rounding=8)
    (49.76681683, -7.55714701)

    But beware that this only works for the area immediately around the British Isles.

    """
    if model not in ELLIPSOID_MODELS:
        raise UndefinedModelError(model)

    if northing is None:
        if type(easting) is tuple:
            (easting, northing) = easting
        else:
            raise MissingArgumentError(easting)

    def _appd(input):
        "Find appropriate decimals for this coordinate"
        s = '{:.3f}'.format(input)
        return 6 if s.endswith('.000') else 9

    decimals = rounding if type(rounding) is int else max(_appd(easting), _appd(northing))

    (os_lat, os_lon) = _reverse_project_onto_ellipsoid(easting, northing, 'OSGB36')

    # if we want OS map LL we are done
    if model == 'OSGB36':
        return (round(os_lat, decimals), round(os_lon, decimals))

    # If we want WGS84 LL, we must adjust to pseudo grid if we can
    shifts = _find_OSTN_shifts_at(easting, northing)
    if shifts is not None:
        in_ostn_polygon = True
        x = easting - shifts[0]
        y = northing - shifts[1]
        last_shifts = shifts[:]
        for _ in range(20):
            shifts = _find_OSTN_shifts_at(x, y)

            if shifts is None:
                # we have been shifted off the edge
                in_ostn_polygon = False
                break

            x = easting - shifts[0]
            y = northing - shifts[1]
            if abs(shifts[0] - last_shifts[0]) < 0.0001:
                if abs(shifts[1] - last_shifts[1]) < 0.0001:
                    break

            last_shifts = shifts[:]

        if in_ostn_polygon:
            return tuple(round(x, decimals) for x in _reverse_project_onto_ellipsoid(x, y, 'WGS84'))

    # If we get here, we must use the Helmert approx
    return tuple(round(x, decimals) for x in _shift_ll_from_osgb36_to_wgs84(os_lat, os_lon))


def ll_to_grid(lat, lon, model='WGS84', rounding=None):
    """Convert a (latitude, longitude) pair to an OSGB grid (easting, northing) pair.

    Output
        a tuple containing (easting, northing) in metres from the grid origin.

    Input
         The arguments should be supplied as real numbers representing
         decimal degrees, like this:

        >>> ll_to_grid(51.5, -2.1)
        (393154.813, 177900.607)

        Following the normal convention, positive arguments mean North
        or East, negative South or West.

        If you have data with degrees, minutes and seconds, you can
        convert them to decimals like this:

        >>> ll_to_grid(51+25/60, 0-5/60-2/3600)
        (533338.156, 170369.238)

        >>> ll_to_grid(52 + 39/60 + 27.2531/3600, 1 + 43/60 + 4.5177/3600, model='OSGB36')
        (651409.903, 313177.27)

        But if you are still using python2 then be sure to ``import
        division`` so that you get the correct semantics for division
        when both numerator and denominator are integers.

    If you have trouble remembering the order of the arguments, or the
    returned values, note that latitude comes before longitude in the
    alphabet too, as easting comes before northing.  However since
    reasonable latitudes for the OSGB are in the range 49 to 61, and
    reasonable longitudes in the range -9 to +2, the ``ll_to_grid``
    function accepts argument in either order.  If your longitude is
    larger than your latitude, then the values of the arguments will be
    silently swapped:

    >>> ll_to_grid(-2.1, 51.5)
    (393154.813, 177900.607)

    But you can always give the arguments as named keywords if you prefer:

    >>> ll_to_grid(lon=-2.1, lat=51.5)
    (393154.813, 177900.607)

    The easting and northing will be returned as the distance in metres
    from the 'false point of origin' of the British Grid (which is a
    point some way to the south-west of the Scilly Isles).  If you want
    the result presented in a more traditional grid reference format you
    should pass the results to :py:func:`format_grid` from
    :py:mod:`osgb.gridder`.

    If the coordinates you supply are in the area covered by the OSTN
    transformation data, then the results will be rounded to 3 decimal
    places, which corresponds to the nearest millimetre.  If they are
    outside the coverage then the conversion is automagically done using
    a Helmert transformation instead of the OSTN data.  The results will
    be rounded to the nearest metre in this case, although you probably
    should not rely on the results being more accurate than about 5m.

    >>> # Somewhere in London
    >>> ll_to_grid(51.3, 0)
    (539524.836, 157551.913)

    >>> # Far north
    >>> ll_to_grid(61.3, 0)
    (507242.0, 1270342.0)

    The coverage extends quite a long way off shore.

    >>> # A point in the sea, to the north-west of Coll
    >>> ll_to_grid(56.75, -7)
    (94469.613, 773209.471)

    The numbers returned may be negative if your latitude and longitude
    are far enough south and west, but beware that the transformation is
    less and less accurate or useful the further you get from the
    British Isles.

    >>> ll_to_grid(51.3, -10)
    (-157250.0, 186110.0)

    ``ll_to_grid`` also takes an optional argument that sets the ellipsoid
    model to use.  This defaults to ``WGS84``, the name of the normal model for
    working with normal GPS coordinates, but if you want to work with the
    traditional latitude and longitude values printed on OS maps before 2015
    then you should add an optional model argument

    >>> ll_to_grid(49, -2, model='OSGB36')
    (400000.0, -100000.0)


    Incidentally, the grid coordinates returned by this call are the
    coordinates of the 'true point of origin' of the British grid.  You
    should get back an easting of 400000.0 for any point with longitude
    2W since this is the central meridian used for the OSGB projection.
    However you will get a slightly different value unless you specify
    ``model='OSGB36'`` since the WGS84 meridians are not quite the same
    as OSGB36.

        >>> ll_to_grid(52, -2, model='OSGB36')
        (400000.0, 233553.731)

        >>> ll_to_grid(52, -2, model='WGS84')
        (400096.274, 233505.403)

    If the model is not ``OSGB36`` or ``WGS84`` you will get an
    UndefinedModelError exception:

        >>> ll_to_grid(52, -2, model='EDM50') # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        UndefinedModelError: EDM50

    You can also control the rounding directly if you need to, but be
    aware that asking for more decimal places does not make the
    conversion any more accurate; the formulae used are only designed to
    be accurate to 1mm.

        >>> ll_to_grid(52, -2, rounding=4)
        (400096.2738, 233505.4033)

    The allowed values are the same as the second argument
    to the Python built-in ``round`` function.  So you can round to
    lower precision if you want.

        >>> ll_to_grid(52, -2, rounding=-2)
        (400100.0, 233500.0)

    Notice that the Python BIF ``round`` is used to do the rounding here, so
    the grid coordinates will be given to the nearest metre or mm or whatever
    size you choose.  This follows the sample data given by the OS, but it is
    slightly at odds with the OSGB National Grid convention that a given point
    is represented by the coordinates of the SW corner of the square that
    encloses it.  So your best strategy is to use the default rounding here,
    and rely on ``osgb.gridder.format_grid`` to do the *correct* rounding down
    when it formats the coordinates into a grid reference string.

    """

    if lat < lon:
        (lat, lon) = (lon, lat)

    if model not in ELLIPSOID_MODELS:
        raise UndefinedModelError(model)

    easting, northing = _project_onto_grid(lat, lon, model)

    default_decimals = 3
    if model == 'WGS84':
        shifts = _find_OSTN_shifts_at(easting, northing)
        if shifts is not None:
            easting += shifts[0]
            northing += shifts[1]
        else:
            (osgb_lat, osgb_lon) = _shift_ll_from_wgs84_to_osgb36(lat, lon)
            (easting, northing) = _project_onto_grid(osgb_lat, osgb_lon, 'OSGB36')
            default_decimals = 0

    decimals = rounding if type(rounding) is int else default_decimals
    return (round(easting, decimals), round(northing, decimals))


def _compute_M(phi, model):
    '''Compute the first term of the solution given phi.

    Uses the ellipsoid constants (so needs the model name)
    '''
    p_plus = phi + ORIGIN_PHI
    p_minus = phi - ORIGIN_PHI
    _, b, n, _ = ELLIPSOID_MODELS[model]

    return CONVERGENCE_FACTOR * b * (
        (1 + n * (1 + 5 / 4 * n * (1 + n))) * p_minus
        - 3 * n * (1 + n * (1 + 7 / 8 * n)) * math.sin(p_minus) * math.cos(p_plus)
        + (15 / 8 * n * (n * (1 + n))) * math.sin(2 * p_minus) * math.cos(2 * p_plus)
        - 35 / 24 * n * n * n * math.sin(3 * p_minus) * math.cos(3 * p_plus)
    )


def _project_onto_grid(lat, lon, model):
    '''Project spherical coordinates (lat, lon) onto a flat grid.

    This is the core bit of arithmetic, following the OSGB reference
    implementation.  The strange variable names (I, II, III, etc) follow
    those used in the OSGB notes, except that M is used instead of I to
    keep flake8 happy.

    We are essentially using a Taylor polynomial expansion, and the
    accuracy of this projection is limited by the number of terms
    included.  The design point appears to be approximately 1mm of
    error, but this is never precisely defined in the OSGB notes.

    This routine is not meant to be called by the user.  Use
    ll_to_grid() instead.

    >>> _project_onto_grid(52, -2, 'OSGB36')
    (400000.0, 233553.73133031745)

    '''

    phi = lat / 57.29577951308232087679815481410517
    cp = math.cos(phi)
    sp = math.sin(phi)
    tp = sp / cp  # cos phi cannot be zero in GB

    a, _, _, e2 = ELLIPSOID_MODELS[model]

    M = _compute_M(phi, model)

    nu = CONVERGENCE_FACTOR * a / math.sqrt(1 - e2 * sp * sp)
    etasq = (1 - e2 * sp * sp) / (1 - e2) - 1

    II = nu / 2 * sp * cp
    III = nu / 24 * sp * cp**3 * (5 - tp * tp + 9 * etasq)
    IIIA = nu / 720 * sp * cp**5 * (61 + (-58 + tp * tp) * tp * tp)

    IV = nu * cp
    V = nu / 6 * cp**3 * (etasq + 1 - tp * tp)
    VI = nu / 120 * cp**5 * (5 + (-18 + tp * tp) * tp * tp + 14 * etasq - 58 * tp * tp * etasq)

    dl = lon / 57.29577951308232087679815481410517 - ORIGIN_LAMBDA
    north = ORIGIN_NORTHING + M + (II + (III + IIIA * dl * dl) * dl * dl) * dl * dl
    east = ORIGIN_EASTING + (IV + (V + VI * dl * dl) * dl * dl) * dl

    # return them with easting first
    return (east, north)


def _reverse_project_onto_ellipsoid(easting, northing, model):
    '''Un-project from the grid plane back on to the globe.

    This is the core arithmetic for the reverse projection.  Don't use
    this directly.  Use grid_to_ll instead.

    The strange variable names follow (roughly) the OSGB formulae.  The
    accuracy is limited by the number of terms in the final expansions,
    and the speed by the fact that this has to be an iterative process.

    >>> _reverse_project_onto_ellipsoid(400000.0, 233553.731330343, 'OSGB36')
    (52.0, -2.0)

    >>> (lat, lon) = _reverse_project_onto_ellipsoid(651409.903, 313177.270, 'OSGB36')
    >>> print('{:.8f} {:.8f}'.format(lat, lon))
    52.65757030 1.71792158

    '''

    _, _, _, e2 = ELLIPSOID_MODELS[model]

    af = CONVERGENCE_FACTOR * ELLIPSOID_MODELS[model][0]

    dn = northing - ORIGIN_NORTHING
    de = easting - ORIGIN_EASTING

    phi = ORIGIN_PHI + dn / af

    while True:
        M = _compute_M(phi, model)
        if abs(dn - M) < 0.00001:  # HUNDREDTH_MM
            break
        phi = phi + (dn - M) / af

    cp = math.cos(phi)
    sp = math.sin(phi)
    tp = sp / cp  # math.cos phi cannot be zero in GB

    splat = 1 - e2 * sp * sp
    sqrtsplat = math.sqrt(splat)
    nu = af / sqrtsplat
    rho = af * (1 - e2) / (splat * sqrtsplat)
    etasq = nu / rho - 1

    VII = tp / (2 * rho * nu)
    VIII = (5 + 3 * tp * tp + etasq - 9 * tp * tp * etasq) * tp / (24 * rho * nu**3)
    IX = (61 + (90 + 45 * tp * tp) * tp * tp) * tp / (720 * rho * nu**5)

    secp = 1 / cp

    X = secp / nu
    XI = secp / (6 * nu**3) * (nu / rho + 2 * tp * tp)
    XII = secp / (120 * nu**5) * (5 + (28 + 24 * tp * tp) * tp * tp)
    XIIA = secp / (5040 * nu**7) * (61 + (662 + (1320 + 720 * tp * tp) * tp * tp) * tp * tp)

    phi = phi + (-VII + (VIII - IX * de * de) * de * de) * de * de
    lam = ORIGIN_LAMBDA + (X + (-XI + (XII - XIIA * de * de) * de * de) * de * de) * de

    # now put into degrees & return
    return (phi * 57.29577951308232087679815481410517,
            lam * 57.29577951308232087679815481410517)


def _km_parts(metres):
    '''Take a positive length in metres and return whole km, plus fractional km part

    >>> _km_parts(3.1415)
    (0, 0.0031415)

    >>> _km_parts(543245.345)
    (543, 0.245345)

    >>> _km_parts(0)
    (0, 0.0)

    >>> _km_parts(1241244)
    (1241, 0.244)

    >>> _km_parts(50000)
    (50, 0.0)

    '''
    x, y = divmod(metres / 1000, 1)
    return (int(x), round(y, 9))


def _find_OSTN_shifts_at(easting, northing):
    '''Get the OSTN shifted at a pseudo grid reference.

    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(331439.160, 431992.943)))
    95.40442 -72.14955

    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(91400.00044, 11399.99932))) # TP01
    92.14556 -81.19532
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(170277.18937, 11652.89486))) # TP02
    93.52863 -80.48986
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(250265.78908, 62095.88359))) # TP03
    94.02192 -79.31459
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(449719.40261, 75415.5941))) # TP04
    96.96839 -79.73310
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(438614.04492, 114871.19188))) # TP05
    96.87508 -78.94188
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(292090.28885, 168081.28118))) # TP06
    94.58115 -77.81618
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(639720.2239, 169645.8238))) # TP07
    101.61110 -79.96580
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(362174.40797, 170056.49988))) # TP08
    95.58303 -77.80988
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(530526.41231, 178467.04377))) # TP09
    98.56169 -78.57977
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(241030.73082, 220409.85756))) # TP10
    93.85318 -77.21656
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(599345.19593, 225801.4851))) # TP11
    100.39407 -78.65910
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(389448.04202, 261989.2714))) # TP12
    96.14798 -77.11840
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(474237.87392, 262125.33306))) # TP13
    98.09508 -77.57806
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(562079.80467, 319862.04179))) # TP14
    100.74233 -77.04679
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(453904.2695, 340910.74271))) # TP15
    98.56450 -75.79971
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(357359.68298, 383364.15112))) # TP16
    96.16002 -73.71512
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(247865.44739, 393566.26455))) # TP17
    93.52361 -73.35555
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(247865.71739, 393568.93846))) # TP18
    93.52361 -73.35546
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(331439.15958, 431992.94355))) # TP19
    95.40442 -72.14955
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(422143.67886, 433891.20633))) # TP20
    98.50714 -72.50533
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(227685.88237, 468918.33093))) # TP21
    92.44763 -70.94293
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(525643.49032, 470775.61009))) # TP22
    102.17968 -72.39609
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(244687.5168, 495324.61116))) # TP23
    93.11920 -69.72416
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(339824.59869, 556102.50434))) # TP24
    96.54631 -67.74334
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(424539.71821, 565080.53254))) # TP25
    99.63679 -67.82954
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(256247.4854, 664760.29198))) # TP26
    93.43960 -63.02298
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(319092.32821, 671010.63051))) # TP27
    96.10579 -63.09651
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(167542.80567, 797124.1528))) # TP28
    91.39633 -57.00880
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(397061.06978, 805408.1455))) # TP29
    99.42122 -58.40950
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(266961.48086, 846233.64636))) # TP30
    95.28714 -56.67436
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(9500.0055, 899499.9915))) # TP31
    87.90350 -50.99550
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(71622.45111, 938567.30194))) # TP32
    90.68089 -50.89794
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(151874.98359, 966535.33103))) # TP33
    93.66841 -51.55103
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(299624.62633, 967256.59533))) # TP34
    97.26467 -53.60333
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(330299.99994, 1017400.00024))) # TP35
    98.32306 -52.98424
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(261499.99975, 1025500.00025))) # TP36
    96.77825 -52.39825
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(180766.8244, 1029654.63912))) # TP37
    95.63660 -50.52512
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(421199.99984, 1072200.0002))) # TP38
    100.52516 -52.76120
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(440623.59268, 1107930.78086))) # TP39
    101.48032 -52.33286
    >>> print("{:.5f} {:.5f}".format(*_find_OSTN_shifts_at(395898.57806, 1138780.34555))) # TP40
    101.08994 -51.39455
    '''

    if not 0 < easting < 700000:
        return None

    if not 0 < northing < 1250000:
        return None

    east_km, t = _km_parts(easting)
    north_km, u = _km_parts(northing)

    lle = (OSTN_EE_BASE + OSTN_EE_SHIFTS[east_km + north_km * 701]) / 1000
    lre = (OSTN_EE_BASE + OSTN_EE_SHIFTS[east_km + north_km * 701 + 1]) / 1000
    ule = (OSTN_EE_BASE + OSTN_EE_SHIFTS[east_km + north_km * 701 + 701]) / 1000
    ure = (OSTN_EE_BASE + OSTN_EE_SHIFTS[east_km + north_km * 701 + 702]) / 1000

    lln = (OSTN_NN_BASE + OSTN_NN_SHIFTS[east_km + north_km * 701]) / 1000
    lrn = (OSTN_NN_BASE + OSTN_NN_SHIFTS[east_km + north_km * 701 + 1]) / 1000
    uln = (OSTN_NN_BASE + OSTN_NN_SHIFTS[east_km + north_km * 701 + 701]) / 1000
    urn = (OSTN_NN_BASE + OSTN_NN_SHIFTS[east_km + north_km * 701 + 702]) / 1000

    return (
        (1 - t) * (1 - u) * lle + t * (1 - u) * lre + (1 - t) * u * ule + t * u * ure,
        (1 - t) * (1 - u) * lln + t * (1 - u) * lrn + (1 - t) * u * uln + t * u * urn
    )


def _llh_to_cartesian(lat, lon, H, model):
    '''Approximate conversion from spherical to plane coordinates.

    Used as part of the Helmert transformation used outside the OSTN02
    area.

    >>> _llh_to_cartesian(53, -3, 10, 'OSGB36')
    (3841039.2016489906, -201300.3346975291, 5070178.453880735)

    >>> (x, y, z) =  _llh_to_cartesian(52, 1, 30, 'WGS84')
    >>> tuple(round(x, 8) for x in _cartesian_to_llh(x, y, z, 'WGS84'))
    (52.0, 1.0, 30.0)

    Note that you dont get more that 8 places of precision here.  Hence
    the `round(x, 8)` in the test. But that gives you precision to
    within 1mm which was the design point originally chosen by the OSGB.

    Numbers from the worked example in the OSGB guide

    e2 == 6.6705397616E-03
    nu == 6.3910506260E+06

    >>> t = _llh_to_cartesian(52 + 39/60 + 27.2531/3600, 1 + 43/60 + 4.5177/3600, 24.700, 'OSGB36')
    >>> tuple(round(x, 4) for x in t)
    (3874938.8497, 116218.6238, 5047168.2073)

    Note that here we have rounded to 4 places because these are in
    units of metres rather than degrees.

    '''
    a, _, _, e2 = ELLIPSOID_MODELS[model]

    phi = lat / 57.29577951308232087679815481410517
    sp = math.sin(phi)
    cp = math.cos(phi)

    lam = lon / 57.29577951308232087679815481410517
    sl = math.sin(lam)
    cl = math.cos(lam)

    nu = a / math.sqrt(1 - e2 * sp * sp)

    return (
        (nu + H) * cp * cl,
        (nu + H) * cp * sl,
        ((1 - e2) * nu + H) * sp
    )


def _cartesian_to_llh(x, y, z, model):
    '''Approximate conversion from plane to spherical coordinates.

    Used as part of the Helmert transformation used outside the OSTN02
    area.

    >>> _cartesian_to_llh(3841039.2016489909, -201300.3346975291, 5070178.453880735, 'OSGB36')
    (53.0, -3.0, 10.0)
    '''

    a, _, _, e2 = ELLIPSOID_MODELS[model]

    p = math.sqrt(x * x + y * y)
    lam = math.atan2(y, x)
    phi = math.atan2(z, p * (1 - e2))

    while True:
        sp = math.sin(phi)
        nu = a / math.sqrt(1 - e2 * sp * sp)
        oldphi = phi
        phi = math.atan2(z + e2 * nu * sp, p)
        if abs(oldphi - phi) < 1E-12:
            break

    return (
        phi * 57.29577951308232087679815481410517,
        lam * 57.29577951308232087679815481410517,
        p / math.cos(phi) - nu
    )


def _small_Helmert_transform_for_OSGB(direction, xa, ya, za):
    '''Transform 3d planar coordinates to approximate OSGB36 to WGS84.

    Formulae as supplied by the OSGB.  Designed for +/- 5m accuracy in
    most of OSGB area.

    `direction` indicates the desired transformation: -1 -> WGS84, +1 -> OSGB36

    '''
    tx = direction * -446.448
    ty = direction * +125.157
    tz = direction * -542.060
    sp = direction * 0.0000204894 + 1
    rx = (direction * -0.1502 / 3600) / 57.29577951308232087679815481410517
    ry = (direction * -0.2470 / 3600) / 57.29577951308232087679815481410517
    rz = (direction * -0.8421 / 3600) / 57.29577951308232087679815481410517
    xb = tx + sp * xa - rz * ya + ry * za
    yb = ty + rz * xa + sp * ya - rx * za
    zb = tz - ry * xa + rx * ya + sp * za
    return (xb, yb, zb)


def _shift_ll_from_osgb36_to_wgs84(lat, lon):
    '''Approximate conversion of OGSB sperical coordinates to WGS84.

    Used as a last resort by grid_to_ll

    >>> _shift_ll_from_osgb36_to_wgs84(52, -2)
    (52.0004248559296, -2.0014104687899463)
    '''
    (xa, ya, za) = _llh_to_cartesian(lat, lon, 0, 'OSGB36')
    (xb, yb, zb) = _small_Helmert_transform_for_OSGB(-1, xa, ya, za)
    (latx, lonx, _) = _cartesian_to_llh(xb, yb, zb, 'WGS84')
    return (latx, lonx)


def _shift_ll_from_wgs84_to_osgb36(lat, lon):
    '''Approximate conversion of WGS84 spherical coordinates to OSGB36.

    Used as a last resort by ll_to_grid
    >>> tuple(round(x, 7) for x in _shift_ll_from_wgs84_to_osgb36(52.0004248559296, -2.0014104687899463))
    (52.0, -2.0)
    '''
    (xa, ya, za) = _llh_to_cartesian(lat, lon, 0, 'WGS84')
    (xb, yb, zb) = _small_Helmert_transform_for_OSGB(+1, xa, ya, za)
    (latx, lonx, _) = _cartesian_to_llh(xb, yb, zb, 'OSGB36')
    return (latx, lonx)



'''Print a map of the National Grid, optionally with an index of OS maps'''
from __future__ import print_function, division

import argparse
import math
import os
import pkgutil
import subprocess
import sys
import tempfile
import textwrap

import osgb


def does_not_overlap_parent(key):
    "See if an inset overlaps the parent sheet"
    inset = osgb.map_locker[key]
    parent = osgb.map_locker[inset.parent]
    return (inset.bbox[0][0] > parent.bbox[1][0]
            or inset.bbox[1][0] < parent.bbox[0][0]
            or inset.bbox[1][1] < parent.bbox[0][1]
            or inset.bbox[0][1] > parent.bbox[1][1])


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog='plot_maps',
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog="Toby Thurston -- August 2021",
                                     description=textwrap.dedent('''
    plot_maps - make a nice index sheet for a map series.

    If you have a working TeXLive installation with MetaPost and GhostScript installed,
    you can use it to produce PDF index maps of the various map series included.'''))

    parser.add_argument('--series', type=str, choices='ABCHJ', help='Show map series')
    parser.add_argument('--paper', type=str.upper, choices=('A0', 'A1', 'A2', 'A3', 'A4'),
                        default='A3', help='Output paper size')
    parser.add_argument('--pdf', type=str, help="Write to this file")
    parser.add_argument('--towns', action='store_true', help="Show some town names")
    parser.add_argument('--tests', action='store_true', help="Show the OSGB test locations")
    parser.add_argument('--error', action='store_true', help="Show an indication of the round trip error")
    parser.add_argument('--nogrid', action='store_true', help="Don't show the grid")
    parser.add_argument('--nograt', action='store_true', help="Don't show the latitude / longitude graticule")
    parser.add_argument('--nocoast', action='store_true', help="Don't show the coast lines")
    parser.add_argument('--nomp', action='store_true', help="Don't try to run MP")
    args = parser.parse_args()

    # Where shall we put the output?
    if args.pdf:
        pdffile = args.pdf
    else:
        if args.series:
            pdffile = "Index_for_map_series_{}.pdf".format(args.series)
        else:
            pdffile = "National_grid.pdf"

    # Set the scale according to the chosen paper size
    scale = {'A4': 1680, 'A3': 1189, 'A2': 840, 'A1': 597, 'A0': 420}[args.paper]

    # gather all the paths we will need from the map locker polygons
    path_for = dict()  # a list of the actual MP maps
    sides = list()  # a list of keys to path_for
    insets = list()  # another list of (different) keys to path_for
    if args.series:
        for k, sheet in osgb.map_locker.items():
            # Skip maps not wanted
            if k[:1] not in args.series:
                continue

            # make the polygon into an MP path
            path_for[k] = '--'.join('({:.1f}, {:.1f})'.format(x[0]/scale, x[1]/scale)
                                    for x in sheet.polygon[:-1]) + '--cycle'
            # append the key to the appropriate list
            if sheet.parent == '':
                sides.append(k)
            else:
                insets.append(k)

    # MP rgbcolor tuples
    color_for = {
        'A': '(224/255, 36/255, 114/255)',  # Landranger pink
        'B': '(221/255, 61/255, 31/255)',  # Explorer orange
        'C': '(228/255, 0, 28/255)',  # Seventh series red
        'H': '(128/255, 4/255, 36/255)',  # Harvey dark red
        'J': '(128/255, 4/255, 36/255)',  # Harvey dark red
    }

    # open a tempory file for MP
    plotter = tempfile.NamedTemporaryFile(mode='wt', prefix='plot_maps_', suffix='.mp', dir='.', delete=False)

    # Starting making the MP file
    print('prologues := 3; outputtemplate := "%j.eps"; beginfig(1); defaultfont := "phvr8r";', file=plotter)

    # sides and insets will have anything in if we chose one or more series
    for k in sides + insets:
        print("fill {} withcolor (0.98, 0.906, 0.71);".format(path_for[k]), file=plotter)

    if args.error:
        for x in range(70):
            e = x * 10000 + 5000
            for y in range(125):
                n = y * 10000 + 5000
                (ee, nn) = osgb.ll_to_grid(*osgb.grid_to_ll(e, n))
                h = math.sqrt((e-ee)**2+(n-nn)**2)
                if h > 0:
                    print('drawdot ({:.1f}, {:.1f})'.format(e/scale, n/scale), file=plotter)
                    print(' withpen pencircle scaled 4 withcolor {:.2f}[white, red];'.format(h*100), file=plotter)

        print('label.rt("Round trip error (mm)" infont defaultfont scaled 0.6,', file=plotter)
        print('({:.1f}, {:.1f}));'.format(-176500/scale, 1255000/scale), file=plotter)
        for i in range(6):
            e = -120000 - i * 10000
            n = 1245000
            print('drawdot ({:.1f}, {:.1f})'.format(e/scale, n/scale), file=plotter)
            print(' withpen pencircle scaled 4 withcolor {:.2f}[white, red];'.format(i/5), file=plotter)
            print('label.bot("{}" infont defaultfont scaled 0.6,'.format(2*i), file=plotter)
            print('({:.1f}, {:.1f}));'.format(e/scale, n/scale-2), file=plotter)

    if not args.nograt:
        print("drawoptions(withpen pencircle scaled 0.4);", file=plotter)
        for lon in range(-10, 3):
            points = []
            for decilat in range(496, 613):
                e, n = osgb.ll_to_grid(decilat/10, lon)
                points.append('({:.1f}, {:.1f})'.format(e/scale, n/scale))

            print('draw ' + '--'.join(points) + ' withcolor .7[.5 green, white];', file=plotter)
            print('label.bot("{}" & char 176, {}) withcolor .4 green;'.format(lon, points[0]), file=plotter)

        for lat in range(50, 62):
            points = []
            for decilon in range(-102, 23):
                e, n = osgb.ll_to_grid(lat, decilon/10)
                points.append('({:.1f}, {:.1f})'.format(e/scale, n/scale))

            print('draw ' + '--'.join(points) + ' withcolor .7[.5 green, white];', file=plotter)
            print('label.lft("{}" & char 176, {}) withcolor .4 green;'.format(lat, points[0]), file=plotter)

    if not args.nogrid:
        print('drawoptions(withcolor .7 white);', file=plotter)
        print('z0=({:g}, {:g});'.format(700000/scale, 1250000/scale), file=plotter)
        print('label.llft("0", origin) withcolor .5 white;', file=plotter)

        for i in range(8):
            e = i*100000
            print('t:={:g};draw (t, 0) -- (t, y0);'.format(e/scale), file=plotter)
            if i > 0:
                print('label.bot("{:d}", (t, 0)) withcolor .5 white;'.format(i*100), file=plotter)
            for j in range(13):
                n = j*100000
                if i == 0:
                    print('t:={:g};draw (0, t) -- (x0, t);'.format(n/scale), file=plotter)
                    if j > 0:
                        print('label.lft("{:d}", (0, t)) withcolor .5 white;'.format(j*100), file=plotter)

                if i < 7 and j < 12:
                    sq = osgb.format_grid(e, n, form='SS')
                    print('label("{}" infont "phvr8r" scaled {},'.format(sq, 3600/scale), file=plotter)
                    print('({:.1f}, {:.1f})) withcolor 3/4;'.format((e+50000)/scale, (n+50000)/scale), file=plotter)
        # add HP as well
        print('label("HP" infont "phvr8r" scaled {},'.format(3600/scale), file=plotter)
        print('({:.1f}, {:.1f})) withcolor 3/4;'.format((450000)/scale, (1250000)/scale), file=plotter)

    if not args.nocoast:
        coast_shapes = pkgutil.get_data('osgb', 'gb_coastline.shapes')
        if coast_shapes:
            print("drawoptions(withpen pencircle scaled 0.2 withcolor (0, 172/255, 226/255));", file=plotter)
            poly_path = list()
            for line in coast_shapes.split(b'\n'):
                if line.startswith(b'#'):
                    print('draw ' + '--'.join(poly_path) + ';', file=plotter)
                    del poly_path[:]
                elif line:
                    try:
                        (lon, lat) = (float(x) for x in line.split())
                    except ValueError:
                        print('????', line)
                    (e, n) = osgb.ll_to_grid(lat, lon)
                    poly_path.append('({:.1f}, {:.1f})'.format(e/scale, n/scale))

            assert not poly_path

    if args.towns:
        towns = {
            'Aberdeen': (392500, 806500),
            'Birmingham': (409500, 287500),
            'Bristol': (360500, 175500),
            'Cambridge': (546500, 258500),
            'Canterbury': (614500, 157500),
            'Cardiff': (318500, 176500),
            'Carlisle': (339500, 555500),
            'Edinburgh': (327500, 673500),
            'Glasgow': (259500, 665500),
            'Inverness': (266500, 845500),
            'Leeds': (430500, 434500),
            'Liverpool': (337500, 391500),
            'London': (531500, 181500),
            'Manchester': (383500, 398500),
            'Newcastle': (425500, 564500),
            'Oxford': (451500, 206500),
            'Plymouth': (247500, 56500),
            'Portsmouth': (465500, 101500),
            'Salisbury': (414500, 130500),
            'Sheffield': (435500, 387500),
            'Worcester': (385500, 255500),
        }

        print("drawoptions(withcolor .7 white);defaultscale := 1/2;", file=plotter)
        for t in towns:
            e, n = towns[t]
            print('dotlabel.top("{}", ({:.1f}, {:.1f}));'.format(t, e/scale, n/scale), file=plotter)

    if args.tests:
        points = {
            'TP01': (91492.146, 11318.803, "St Mary's, Scilly"),
            'TP02': (170370.718, 11572.405, "Lizard Point Lighthouse"),
            'TP03': (250359.811, 62016.569, "Plymouth"),
            'TP04': (449816.371, 75335.861, "St Catherine's Point Lighthouse"),
            'TP05': (438710.92, 114792.25, "Former OSHQ"),
            'TP06': (292184.87, 168003.465, "Nash Point Lighthouse"),
            'TP07': (639821.835, 169565.858, "North Foreland Lighthouse"),
            'TP08': (362269.991, 169978.69, "Brislington"),
            'TP09': (530624.974, 178388.464, "Lambeth"),
            'TP10': (241124.584, 220332.641, "Carmarthen"),
            'TP11': (599445.59, 225722.826, "Colchester"),
            'TP12': (389544.19, 261912.153, "Droitwich"),
            'TP13': (474335.969, 262047.755, "Northampton"),
            'TP14': (562180.547, 319784.995, "King's Lynn"),
            'TP15': (454002.834, 340834.943, "Nottingham"),
            'TP16': (357455.843, 383290.436, "STFC, Daresbury"),
            'TP17': (247958.971, 393492.909, "Point Lynas Lighthouse, Anglesey"),
            'TP18': (247959.241, 393495.583, "Point Lynas Lighthouse, Anglesey"),
            'TP19': (331534.564, 431920.794, "Blackpool Airport"),
            'TP20': (422242.186, 433818.701, "Pudsey"),
            'TP21': (227778.33, 468847.388, "Isle of Man airport"),
            'TP22': (525745.67, 470703.214, "Flamborough Head"),
            'TP23': (244780.636, 495254.887, "Ramsey, Isle of Man"),
            'TP24': (339921.145, 556034.761, "Carlisle"),
            'TP25': (424639.355, 565012.703, "Newcastle University"),
            'TP26': (256340.925, 664697.269, "Glasgow"),
            'TP27': (319188.434, 670947.534, "Sighthill, Edinburgh"),
            'TP28': (167634.202, 797067.144, "Mallaig Lifeboat Station"),
            'TP29': (397160.491, 805349.736, "Girdle Ness Lighthouse"),
            'TP30': (267056.768, 846176.972, "Inverness"),
            'TP31': (9587.909, 899448.986, "Hirta, St Kilda"),
            'TP32': (71713.132, 938516.4, "at sea, 7km S of Flannan"),
            'TP33': (151968.652, 966483.779, "Butt of Lewis lighthouse"),
            'TP34': (299721.891, 967202.992, "Dounreay Airfield"),
            'TP35': (330398.323, 1017347.016, "Orkney Mainland"),
            'TP36': (261596.778, 1025447.602, "at sea, 1km NW of Sule Skerry"),
            'TP37': (180862.461, 1029604.114, "at sea, 3km south of Rona"),
            'TP38': (421300.525, 1072147.239, "Fair Isle"),
            'TP39': (440725.073, 1107878.448, "Sumburgh Head"),
            'TP40': (395999.668, 1138728.951, "Foula"),
        }

        print("drawoptions(withcolor .5[red, white]);", file=plotter)
        for t in points:
            e, n, name = points[t]
            print("draw unitsquare shifted -(1/2, 1/2) rotated 45 scaled 3", file=plotter)
            print("shifted ({:.1f}, {:.1f});".format(e/scale, n/scale), file=plotter)

    if args.series and sides:  # sides will be empty if none of the maps matched series_wanted

        print("drawoptions(withpen pencircle scaled 0.2);defaultscale:={:.2f};".format(666/scale), file=plotter)

        for k in sides:
            series = k[:1]
            map_color = color_for[series] if series in color_for else 'black'
            print("draw {} withcolor {};".format(path_for[k], map_color), file=plotter)

            sheet = osgb.map_locker[k]
            x = (sheet.bbox[0][0] + sheet.bbox[1][0]) / 2 / scale
            y = (sheet.bbox[0][1] + sheet.bbox[1][1]) / 2 / scale

            if sheet.number.startswith('OL'):
                map_color = '(.5, .5, 1)'
                y += 3

            print('label("{}", ({}, {})) withcolor .76[white, {}];'.format(sheet.number, x, y, map_color), file=plotter)

        print('path p, q;', file=plotter)
        for k in insets:
            series = k[:1]
            map_color = color_for[series] if series in color_for else 'black'
            print("p:={};".format(path_for[k]), file=plotter)
            parent_key = osgb.map_locker[k].parent
            if does_not_overlap_parent(k):
                print('q:={};'.format(path_for[parent_key]), file=plotter)
                print("draw center p -- center q cutbefore p cutafter q", file=plotter)
                print("dashed evenly scaled 1/3 withcolor {};".format(map_color), file=plotter)
            print("draw p withcolor {};".format(map_color), file=plotter)

        y = 1300000/scale
        for s in args.series:
            color = color_for[s] if s in color_for else 'black'
            title = 'label.rt("{} sheet index" infont defaultfont scaled {:.1f}, (0, {:.1f})) withcolor {};'.format(
                    osgb.name_for_map_series[s], 2000/scale, y, color)
            print(title, file=plotter)
            y -= 24

        # add sheet names for Harvey maps
        if args.series in 'HJ':
            print("defaultscale:={:.1f};".format(2000/scale), file=plotter)
            x = 510000/scale
            y = 515000/scale
            print('fill unitsquare xscaled {:.1f}'.format(200000/scale), file=plotter)
            print('yscaled {:.1f}'.format((12000*len(sides)+4000)/scale), file=plotter)
            print('shifted ({:.1f}, {:.1f}) withcolor background;'.format(x-5, y-3), file=plotter)
            for k in sorted(sides, reverse=True):
                sheet = osgb.map_locker[k]
                print('draw "{} {}"'.format(sheet['number'], sheet['title']), file=plotter)
                print('infont defaultfont shifted ({:.1f}, {:.1f});'.format(x, y), file=plotter)
                y += 12000/scale

    # Add a margin
    print('z1 = center currentpicture;', file=plotter)
    print('setbounds currentpicture to bbox currentpicture shifted -z1 scaled 1.05 shifted z1;', file=plotter)

    # Finish the MP input and close the file
    print("endfig;end.", file=plotter)
    plotter.close()

    # Now deal with MP (unless asked not to)
    if not args.nomp:
        try:
            subprocess.check_call(['mpost', plotter.name])
        except subprocess.CalledProcessError as e:
            print('Metapost failed', file=sys.stderr)
            raise e

        epsfile = plotter.name[:-2] + 'eps'
        logfile = plotter.name[:-2] + 'log'

        try:
            subprocess.check_call(['epstopdf', "-o={}".format(pdffile), epsfile])
        except subprocess.CalledProcessError as e:
            print('epdtopdf failed', file=sys.stderr)
            raise e
        try:
            os.unlink(epsfile)
        except OSError as e:
            print("Failed to delete temporary file: {}".format(epsfile), file=sys.stderr)
            raise e

        try:
            os.unlink(logfile)
        except OSError as e:
            print("Failed to delete temporary file: {}".format(logfile), file=sys.stderr)
            raise e

        try:
            os.unlink(plotter.name)
        except OSError as e:
            print('Failed to delete MP file', file=sys.stderr)
            raise e

        print("Created " + pdffile)
    else:
        print("MP source written to " + plotter.name)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""Convert coordinates.

Toby Thurston -- 10 Jun 2022
"""
# allow free variable names and the occasional trailing whitespace character
# pylint: disable=C0103, C0303
from __future__ import print_function, division

import argparse
import random
import re
import webbrowser

import osgb


def get_likely_lon_lat(possible_number):
    "Is this supposed to be a lat or lon coordinate?"
    try:
        actual_number = float(possible_number)
    except ValueError:
        return None

    if -20 < actual_number < 70:  # Useful Lat is between 49 and 65, useful Lon between -9 and 2
        return actual_number  # and it's OK to return 0 - see below

    return None


def hms(degrees):
    '''Sexagesimal notation'''
    sign = '-' if degrees < 0 else ''
    d = abs(degrees)
    dd = int(d)
    m = 60 * (d - dd)
    mm = int(m)
    s = 60 * (m - mm)
    ss = round(s)
    if ss == 60:
        ss = 0
        mm += 1
        if mm == 60:
            dd += 1
            mm = 0
    return "{}{}°{}′{}″".format(sign, dd, mm, ss)


def format_degrees(phi, lam, want_sex=True):
    '''Format degrees as required'''
    if not want_sex:
        return '{:.8f} {:.7f}'.format(phi, lam)

    return hms(phi) + ' ' + hms(lam)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Convert coordinates to and from OSGB grid refs and lat/lon.")
    parser.add_argument("--show", action='store_true', help="Show this place on Streetview.co.uk")
    parser.add_argument("--random", action="store_true", help="Pick a place at random")
    parser.add_argument("--sexagesimal", action="store_true", help="Show degrees in sexagesimal")
    parser.add_argument("grid_or_ll_element", type=str, nargs='*', default=['SP', '101', '203'],
                        help="A grid reference string or lat/lon pair.")
    args = parser.parse_args()

    if args.random:
        map = random.choice(list(S for S in osgb.map_locker.values() if S.series == 'A'))
        e = random.randint(map.bbox[0][0], map.bbox[1][0])
        n = random.randint(map.bbox[0][1], map.bbox[1][1])
        agenda = osgb.format_grid(e, n)

    else:
        agenda = ' '.join(args.grid_or_ll_element)
        alphabet = ' 1234567890-.+/:ABCDEFGHJKLMNOPQRSTUVWXYZabcdef'  # there is no I in this
        agenda = agenda.replace(',', ' ')  # treat commas as spaces
        agenda = ''.join(x for x in agenda if x in alphabet)  # and remove unwanted characters

    m = re.match(r'([456]\d) ([012345]?\d) ([012345]?\d) N (\d) ([012345]?\d) ([012345]?\d) ([EW])', agenda)
    if m is not None:
        (ha, ma, sa, hb, mb, sb, hemi) = m.groups()
        latlon = [int(ha) + int(ma) / 60 + int(sa) / 3600, (int(hb) + int(mb) / 60 + int(sb) / 3600)]
        if hemi == 'W':
            latlon[1] = -latlon[1]
    else:
        latlon = list(get_likely_lon_lat(x) for x in agenda.split())

    # if the arguments consists of just two likely looking numbers...
    # You can't just say "all(floats)" because 0 is a valid Longitude
    if len(latlon) == 2 and all(x is not None for x in latlon):
        (lon, lat) = sorted(latlon)  # Lon is always less than Lat in OSGB
        (e, n) = osgb.ll_to_grid(lat, lon)
        (olat, olon) = (lat, lon)
        (oe, on) = osgb.ll_to_grid(lat, lon, 'OSGB36')

    else:
        print('>>', agenda)
        (e, n) = osgb.parse_grid(agenda)
        (lat, lon) = osgb.grid_to_ll(e, n)
        (oe, on) = (e, n)
        (olat, olon) = osgb.grid_to_ll(e, n, 'OSGB36')

    try:
        grid = osgb.format_grid(e, n)
        maps = osgb.sheet_keys(e, n)
    except osgb.gridder.Error:
        grid = '??'
        maps = None

    if maps:
        map_string = 'on sheets {}'.format(', '.join(maps))
    else:
        map_string = '(not covered by any OSGB map)'

    print('WGS84 ', format_degrees(lat, lon, args.sexagesimal), end=' ')
    print('== {:.3f} {:.3f}'.format(e, n), end=' ')
    print('== {} {}'.format(grid, map_string))

    try:
        grid = osgb.format_grid(oe, on)
        maps = osgb.sheet_keys(oe, on)
    except osgb.gridder.Error:
        grid = '??'
        maps = None

    if maps:
        map_string = 'on sheets {}'.format(', '.join(maps))
    else:
        map_string = '(not covered by any OSGB map)'

    print('OSGB36', format_degrees(olat, olon, args.sexagesimal), end=' ')
    print('== {:.3f} {:.3f}'.format(oe, on), end=' ')
    print('== {} {}'.format(grid, map_string))

    if args.show:
        url = 'http://www.streetmap.co.uk/map.srf?x={:06d}&y={:06d}&z=3'.format(int(e), int(n))
        webbrowser.open(url)



# Compatibility with old osgb interface
import osgb


def lonlat_to_osgb(lon, lat, digits=3, formatted=True, model='OSGB36'):
    """Convert a longitude and latitude to Ordnance Survey grid reference.

    :Parameters:
        lon
            Longitude, assumed to be in OSGB36 degrees
            (unless you set model='WGS84').

        lat
            Latitude, ditto.

        digits
            The number of digits to use for each direction in the final grid reference. Default 3.

        formatted
            Should the OSGB reference be nicely formatted (with whitespace)? Default true.

        model
            'OSGB36' or 'WGS84', default 'OSGB36'

    :Returns:
        A string giving a formatted OSGB reference.

    For example::

        >>> print(lonlat_to_osgb(1.088978, 52.129892))
        TM 114 525
        >>> print(lonlat_to_osgb(1.088978, 52.129892, formatted=False))
        TM114525
        >>> print(lonlat_to_osgb(1.088978, 52.129892, 5))
        TM 11400 52500

    In the re-implemented version you can reverse arguments if you want to...

    >>> print(lonlat_to_osgb(52.129892, 1.088978, 5))
    TM 11400 52500


    """
    east, north = osgb.ll_to_grid(lat, lon, model=model)

    if formatted:
        format_spec = ' '.join(['SS', 'E' * digits, 'N' * digits])
    else:
        format_spec = ''.join(['SS', 'E' * digits, 'N' * digits])

    return osgb.format_grid(east, north, form=format_spec)


def osgb_to_lonlat(osgb_str, model='OSGB36'):
    """
    Convert an Ordinance Survey reference to a longitude and latitude.

    :Parameters:
            osgb_str
                    An Ordnance Survey grid reference in "letter-number" format.
                    Case and spaces are cleaned up by this function, and resolution
                    automatically detected, so that so that ``TM114 525``, ``TM114525``,
                    and ``TM 11400 52500`` are all recognised and identical.

            model
                    'OSGB36' or 'WGS84'.  Default 'OSGB36'.

    :Returns:
            The longitude and latitude of the grid reference, according to the chosen model.

    For example::

            # just outside Ipswich, about 1.088975 52.129892
            >>> osgb_to_lonlat('TM114 525')
            (1.088975, 52.129892)

            # accepts poor formating
            >>> osgb_to_lonlat(' TM 114525 ')
            (1.088975, 52.129892)

            # accepts higher resolution
            >>> osgb_to_lonlat('TM1140052500')
            (1.088975, 52.129892)

    """
    return tuple(reversed(osgb.grid_to_ll(osgb.parse_grid(osgb_str), model=model)))



from osgb.convert import grid_to_ll, ll_to_grid
from osgb.gridder import parse_grid, format_grid, sheet_keys, get_sheet, map_locker, name_for_map_series
from osgb.legacy_interface import lonlat_to_osgb, osgb_to_lonlat



import setuptools

setuptools.setup(
    name='osgb',
    version='1.2.0',
    description='osgb - high-precision geographic coordinate conversion for Great Britain, based on Ordnance Survey data',
    long_description="Python routines for working with grid references as defined by the Ordnance Survey of Great Britain (OSGB).",
    long_description_content_type="text/markdown",
    license='MIT Licence',
    author='Toby Thurston',
    author_email='toby@cpan.org',
    url='http://thurston.eml.cc',
    packages=['osgb'],
    package_data={'osgb': ['ostn_east_shift_82140', 'ostn_north_shift_-84180', 'gb_coastline.shapes',
                           'maps-explorer.txt', 'maps-harvey-mountain.txt', 'maps-harvey-superwalker.txt',
                           'maps-landranger.txt', 'maps-one-inch.txt', ]},
    scripts=['scripts/bngl.py', 'scripts/whatmaps.py', 'scripts/plot_maps.py'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    keywords='GIS geographic coordinates conversion',
)



#! /usr/bin/env python3
"""
Toby Thurston -- 24 Apr 2018

Take a GPX file (route, track, etc), and find all the maps needed to cover it.

"""
# pylint: disable=C0103
from __future__ import division, print_function

import argparse
import collections

import gpxpy
import osgb

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Find all the maps needed to cover a route or track in the GPX file")
    parser.add_argument("gpxfile", help="The name of your GPX file")
    parser.add_argument("--series", default="A",
                        help="Which map series? A: Landranger, B: Explorer, C: One-inch, ...")
    parser.add_argument("--name", action="store_true", help="Show name of the series")
    parser.add_argument("--title", action="store_true", help="Show the title of the map")
    parser.add_argument("--coverage", action="store_true",
                        help="Show how many of the points are covered by each map")

    args = parser.parse_args()

    with open(args.gpxfile) as g:
        gpx = gpxpy.parse(g)

    points = list()
    for track in gpx.tracks:
        for segment in track.segments:
            for point in segment.points:
                points.append(point)

    for point in gpx.waypoints:
        points.append(point)

    for route in gpx.routes:
        for point in route.points:
            points.append(point)

    maps = list()
    for p in points:
        maps.extend(osgb.sheet_keys(*osgb.ll_to_grid(p.latitude, p.longitude), series=args.series))

    c = collections.Counter(maps)
    for map_key in sorted(c):
        out = list()
        if args.name:
            k, sheet = map_key.split(':')
            out.append('{} {}'.format(osgb.name_for_map_series[k], sheet))
        else:
            out.append(map_key)
        if args.title:
            out.append('"' + osgb.map_locker[map_key].title + '"')
        if args.coverage:
            out.append("({}%)".format(int(100*c[map_key]/len(points))))
        print(' '.join(out))
